// Copyright 1994, 2003 by Jon Dart. All Rights Reserved.
// $Revision: 1.4 $
//
#include "movegen.h"
#include "bearing.h"
#include "attacke.h"
#include "util.h"
#include "debug.h"
#include "moveord.h"
#include "config.h"
#include "bearing.h"
#include "globals.h"
#include <iostream>
#include <fstream>

extern const int Direction[2];

Move Killers1[Constants::MaxPly];
Move Killers2[Constants::MaxPly];
static int HistoryW[64][64];
static int HistoryB[64][64];

static inline void swap( Move moves[], const int i, const int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i]; 
   moves[i] = tmp;
}

static int IsValidMove(const Board &board, Move killer) {
  PieceType pieceMoved = PieceMoved(killer);
  Square start = StartSquare(killer); Square dest = DestSquare(killer);
  if (TypeOfPiece(board[start]) == pieceMoved && PieceColor(board[start]) == board.Side()
      && board[dest] == EmptyPiece) {
    switch(pieceMoved) {
     case Knight: 
        return 1;
     case King: {
        if (TypeOfMove(killer) == KCastle) {
           Square kp = board.KingPos(board.Side());
           Board::CastleType CS = board.CastleStatus(board.Side());
           return board[kp] == MakePiece(King,board.Side()) &&
             ((CS == Board::CanCastleEitherSide) ||
             (CS == Board::CanCastleKSide)) &&
             board[kp + 1] == EmptyPiece &&
             board[kp + 2] == EmptyPiece &&
             board.CheckStatus() != Board::InCheck &&
             !board.any_attacks(kp + 1,board.OppositeSide()) &&
             !board.any_attacks(kp + 2,board.OppositeSide());
        }
        else if (TypeOfMove(killer) == QCastle) {
           Square kp = board.KingPos(board.Side());
           Board::CastleType CS = board.CastleStatus(board.Side());
           return board[kp] == MakePiece(King,board.Side()) &&
             ((CS == Board::CanCastleEitherSide) ||
             (CS == Board::CanCastleQSide)) &&
             board[kp - 1] == EmptyPiece &&
             board[kp - 2] == EmptyPiece &&
             board[kp - 3] == EmptyPiece && 
             board.CheckStatus() != Board::InCheck &&
             !board.any_attacks(kp - 1,board.OppositeSide()) &&
             !board.any_attacks(kp - 2,board.OppositeSide());
        }
        else
          return 1;
     }
     case Rook: {
        int d = Util::Abs(Bearing::Directions[start][dest]);
        return (d==1 || d==8) && board.Clear(start,dest);
     }
     case Bishop: {
        int d = Util::Abs(Bearing::Directions[start][dest]);
        return (d==7 || d==9) && board.Clear(start,dest);
     }
     case Queen: {
        int d = Util::Abs(Bearing::Directions[start][dest]);
        return (d==1 || d==8 || d==7 || d==9) && board.Clear(start,dest);
     }
     case Pawn: {
        if (board.Side() == White) {
           if (start-8==dest) return 1;
           else return (board[start-8] == EmptyPiece && Rank(start,White)==2 && start-16==dest);
       }
       else {
           if (start+8==dest) return 1;
           else return (board[start+8] == EmptyPiece && Rank(start,Black)==2 && start+16==dest);
       }
     }
     default:
        break;
     }
  }
  return 0;
}

int Move_Generator::OrderWinningCaptures (const Board &board, Move *moves, int index,
  int captures, Move *losers, int &num_losers, int ply) {
   int j = index;
   int scores[100];
   for (int i = index; i < index+captures; i++) {
      Square start = StartSquare(moves[i]);
      Square dest = DestSquare(moves[i]);
      PieceType capture = TypeOfPiece(board[dest]);
      if (MovesEqual(m_hashMove,moves[i])) {
         SetUsed(moves[i]);
      }
      else if (!IsEmptyPiece(capture))
      {
         int start_val;
         start_val = PieceValue(board[start]);
         int capture_val = PieceValue(capture);
         int gain;
         if (capture_val >= start_val-10 && PromoteTo(moves[i]) == InvalidPiece)
            gain = capture_val - start_val;
         else
            gain = attack_estimate(board,moves[i]);
#ifdef _TRACE
         for (int k=0; k < ply; k++) cout << ' ';
         cout << "move " << MoveImage(moves[i]) << " est = " << gain << endl;
#endif
         if (gain >= -10) {
            if (i != j)
            {
               swap(moves,i,j);
            }
            scores[j] = gain;
            ++j;
         } else {
            losers[num_losers++] = moves[i];
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         int score = PieceValue(PromoteTo(moves[i])) - 64;
#ifdef _TRACE
         for (int k=0; k < ply; k++) cout << ' ';
         cout << "move " << MoveImage(moves[i]) << " est = " << score << endl;
#endif
         scores[j] = score;
         swap(moves,i,j);
         ++j;
      }
      else if (TypeOfMove(moves[i]) == EnPassant) {
         int gain = attack_estimate(board,moves[i]);
         if (gain >= 0) {
            scores[j] = gain;
            swap(moves,i,j);
            ++j;
         }
         else
            losers[num_losers++] = moves[i];   
      }
   }
   int NumMoves = j-index;
   if (NumMoves > 1) {
       Move_Ordering::sort_moves(moves,scores,NumMoves);
   }
   return NumMoves;
}

int Move_Generator::num_evasions() {
   if (phase == 0) {
        evasion_count = Check_Evasions(moves,0);
        ++phase;
   }
   return evasion_count;
}

Move Move_Generator::NextEvasion() {
      switch(phase) {
      case 0: {
        evasion_count = Check_Evasions(moves,0);
        ++phase;
      }
      case 1: {
        // do pv first
        for (int i = 0; i < evasion_count; i++) {
           if (MovesEqual(m_hashMove,moves[i])) {
              if (i!=0) swap(moves,i,0);
              batch_count = 1;
              ++phase;
              return m_hashMove;
           }
        }
        ++phase;
      }
      case 2: {
        for (int i = 0; i < evasion_count; i++) 
           if (MovesEqual(m_hashMove,moves[i])) {
              SetUsed(moves[i]); 
              break;
           }
        batch_count = evasion_count;
        index = 0;
        ++phase;
      }
      default : break;
      }
      if (index < batch_count)
          return moves[index++];
      else
         return NullMove;
}

int Move_Generator::getBatch(Move *&batch,int &index)
{
   unsigned NumMoves  = 0;
   batch = moves;
   for (; NumMoves == 0 && phase<=LOSERS_PHASE; ++phase)
   {
#ifdef _TRACE
      indent(my_ply); cout << "phase = " << phase << endl;
#endif     
      switch(phase) {
      case HASH_MOVE_PHASE: {
         if (!IsNull(m_hashMove)) {
            *moves = m_hashMove;
            index = 0;
            NumMoves = 1;
         }
         break;
      }
      case WINNING_CAPTURE_PHASE: {
         int captures = Move_Generator::Generate_Captures(moves);
         index = 0;
         NumMoves = OrderWinningCaptures(board,moves,index,captures,losers,
           losers_count,my_ply);
         break;
      }
      case KILLER1_PHASE: {
         if (!IsNull(Killers1[my_ply]) && !MovesEqual(m_hashMove,Killers1[my_ply])) {
             if (IsValidMove(board,Killers1[my_ply])) {
                moves[NumMoves++] = Killers1[my_ply];
                index = 0;
            }
	     }
         break;
      }
      case KILLER2_PHASE: {
         if (!IsNull(Killers2[my_ply]) && !MovesEqual(m_hashMove,Killers2[my_ply])) {
             if (IsValidMove(board,Killers2[my_ply])) {
                moves[NumMoves++] = Killers2[my_ply];
                index = 0;
            }
	     }
         break;
      }
      case HISTORY_PHASE: {
         non_captures_count = GenerateNonCaptures(moves);
         index = 0;
         history_count = 0;
         if (1) //options.search.use_history && NumMoves)
         {
            int max = -1, maxi=-1;
            int maxi2 =-1;
            for (int i = 0; i < NumMoves; i++)
            {
               if (MovesEqual(m_hashMove,moves[i])) {
                  SetUsed(moves[i]); 
                  continue;
               }
               else if (MovesEqual(Killers1[my_ply],moves[i]) ||
                        MovesEqual(Killers2[my_ply],moves[i])) {
                  SetUsed(moves[i]); 
                  continue;
               }
	       int score;
               if (board.Side() == White)
               {
                  score = HistoryW[StartSquare(moves[i])][DestSquare(moves[i])];
               }
               else
               {
                  score = HistoryB[StartSquare(moves[i])][DestSquare(moves[i])];
               }
               if (score >= max)
               {
                  if (maxi >=0) {
                     maxi2 = maxi;
                  }
                  maxi = i;
                  max = score;
			   }
           }
           if (maxi != -1) swap(moves,history_count++,maxi);
           if (maxi2 != -1) swap(moves,history_count++,maxi2);
        }
        NumMoves = history_count;
        break;    
      }
      case LOSERS_PHASE: {
         if (losers_count) {
           batch = losers;
         }
         NumMoves = losers_count;
         index = 0;
         break;
      }
      case HISTORY_PHASE2: {
         NumMoves = non_captures_count-history_count;
         index = history_count;
         break;
      }
      } // end switch
   } // end for
   return NumMoves;
}

static void
PawnNonCaptures(const Board & board,Move *moves,int &index)
{
   if (board.Side() == White) {
      Bitmap pawns(board.pawn_bits[White]);
      pawns.shr8();
      pawns.andNot(board.all_occupied);
      pawns.andNot(Bearing::rank_mask[0]);
      Square sq;
      while ((sq = pawns.first_one()) != InvalidSquare) {
         moves[index++] = CreateMove(sq+8,sq,InvalidPiece,Pawn,EmptyPiece,Normal);
         if (Rank(sq,White)==3 && board[sq-8] == EmptyPiece)
           moves[index++] = CreateMove(sq+8,sq-8,InvalidPiece,Pawn,EmptyPiece,Normal);
         pawns.clear(sq);
      }
   }
   else {
      Bitmap pawns(board.pawn_bits[Black]);
      pawns.shl8();
      pawns.andNot(board.all_occupied);
      pawns.andNot(Bearing::rank_mask[7]);
      Square sq;
      while ((sq = pawns.first_one()) != InvalidSquare) {
         moves[index++] = CreateMove(sq-8,sq,InvalidPiece,Pawn,EmptyPiece,Normal);
         if (Rank(sq,Black)==3 && board[sq+8] == EmptyPiece)
           moves[index++] = CreateMove(sq-8,sq+8,InvalidPiece,Pawn,EmptyPiece,Normal);
         pawns.clear(sq);
      }
   }
}


int Move_Generator::GenerateNonCaptures(Move *moves) {
   int NumMoves = 0;
   // castling moves
   const ColorType side = board.Side();
   Board::CastleType CS = board.CastleStatus(side);
   if ((CS == Board::CanCastleEitherSide) ||
       (CS == Board::CanCastleKSide))
   {
      const Square kp = board.KingPos(side);
#ifdef _DEBUG
      if (side == White)
      {
         ASSERT(kp == 60);
         ASSERT(board[kp+3] == whiteRook);
      }
      else
      {
         ASSERT(kp == 4);
         ASSERT(board[kp+3] == blackRook);
      }
#endif
      if (board[kp + 1] == EmptyPiece &&
          board[kp + 2] == EmptyPiece &&
          board.CheckStatus() == Board::NotInCheck &&
          !board.any_attacks(kp + 1,OppositeColor(side)) &&
          !board.any_attacks(kp + 2,OppositeColor(side)))
         // can castle
         moves[NumMoves++] = CreateMove(kp, kp+2, InvalidPiece,
                                        King,EmptyPiece, KCastle);
   }
   if ((CS == Board::CanCastleEitherSide) ||
       (CS == Board::CanCastleQSide))
   {
      const Square kp = board.KingPos(side);
      if (board[kp - 1] == EmptyPiece &&
          board[kp - 2] == EmptyPiece &&
          board[kp - 3] == EmptyPiece && 
          board.CheckStatus() == Board::NotInCheck  &&
          !board.any_attacks(kp - 1,OppositeColor(side)) &&
          !board.any_attacks(kp - 2,OppositeColor(side)))
         // can castle
         moves[NumMoves++] = CreateMove(kp, kp-2, InvalidPiece,
                                        King, EmptyPiece, QCastle);
   }
   // pawn moves
   PawnNonCaptures(board,moves,NumMoves);
   // other moves
   static Square squares[Bearing::MaxBearSq];
   Bitmap pieces(board.occupied[board.Side()]);
   pieces.andNot(board.pawn_bits[board.Side()]);
   Square loc;
   for (;(loc = pieces.first_one()) != InvalidSquare;pieces.clear(loc)) {

       if (TypeOfPiece(board[loc]) == Knight) {
            Bitmap dests(Bearing::knight_attacks[loc]);
            dests.andNot(board.all_occupied);
            Square sq;
            for (;(sq = dests.first_one()) != InvalidSquare;dests.clear(sq))
            {
               moves[NumMoves++] = 
               CreateMove(loc,sq,InvalidPiece,Knight,
                          EmptyPiece,Normal);
            }
            continue;
      }
      else if (TypeOfPiece(board[loc]) == King) {
            Bitmap dests(Bearing::king_attacks[loc]);
            dests.andNot(board.all_occupied);
            Square sq;
            for (;(sq = dests.first_one()) != InvalidSquare;dests.clear(sq))
            {
                moves[NumMoves++] = 
                  CreateMove(loc,sq,InvalidPiece,King,
                        EmptyPiece,Normal);
            }
            continue;
      }
      int n = Bearing::BearSq(board, loc, squares);
      const PieceType piecemoved = TypeOfPiece(board[loc]);
      for (int j = 0; j < n; j++)
      {
         const Square dest = squares[j];
         if (board[dest] == EmptyPiece) {
            moves[NumMoves++] = CreateMove(loc, dest, InvalidPiece,
                                           piecemoved, EmptyPiece, Normal);
         }
      }
   }
   return NumMoves;
}

#define RBAttacks(board, loc, data, moves, NumMoves) \
    Square i; \
    while ((i = *data) != 255 && board[i] == EmptyPiece) \
    { \
       ++data; \
    } \
    if (i != 255 && PieceColor(board[i]) != side) \
       moves[NumMoves++] = \
          CreateMove(loc,i,InvalidPiece,TypeOfPiece(board[loc]), \
          TypeOfPiece(board[i]),Normal) \

int Move_Generator::Generate_Captures(Move * moves)
{
   unsigned NumMoves;
   NumMoves = 0;
   Square loc;

   if (board.Side() == White) {
      Bitmap pawns(board.pawn_bits[White]);
      Bitmap pawns1 = pawns;
      pawns1.shr(7);
      pawns1.andNot(Bitmap(MAKELONGLONG(0x01010101,0x01010101)));
      pawns1.And(board.occupied[Black]);
      Square dest;
      for (;(dest = pawns1.first_one()) != InvalidSquare; pawns1.clear(dest)) {
         Square loc = dest+7;
         int candidate = (Rank(loc,White) == 7);
         if (candidate)
         {
            moves[NumMoves++] =
              CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
              CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
            if (my_ply == 0)
            {
               moves[NumMoves++] =
                 CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
               moves[NumMoves++] =
                 CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
            }
         }
         else
            moves[NumMoves++] =
              CreateMove(loc,dest,InvalidPiece,Pawn,TypeOfPiece(board[dest]),Normal);
      }
      pawns1 = pawns;
      pawns1.shr(9);
      pawns1.andNot(Bitmap(MAKELONGLONG(0x80808080,0x80808080)));
      pawns1.And(board.occupied[Black]);
      for (;(dest = pawns1.first_one()) != InvalidSquare; pawns1.clear(dest)) {
         Square loc = dest+9;
         int candidate = (Rank(loc,White) == 7);
         if (candidate)
         {
            moves[NumMoves++] =
              CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
              CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
            if (my_ply == 0)
            {
               moves[NumMoves++] =
                 CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
               moves[NumMoves++] =
                 CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
            }
         }
         else {
            moves[NumMoves++] =
              CreateMove(loc,dest,InvalidPiece,Pawn,TypeOfPiece(board[dest]),Normal);
         }
      }
      pawns1 = pawns;
      pawns1.And(Bearing::rank_mask[1]);
      pawns1.shr8();
      pawns1.andNot(board.all_occupied);
      for (;(dest = pawns1.first_one()) != InvalidSquare; pawns1.clear(dest)) {
          Square loc = dest + 8;
          moves[NumMoves++] =
            CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
          moves[NumMoves++] =
            CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
          if (my_ply == 0)
          {
             moves[NumMoves++] =
               CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
             moves[NumMoves++] =
               CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
          }
      }
      if (!IsInvalid(board.EnPassantSq()))
      {
         Square epsq = board.EnPassantSq();
         ASSERT(TypeOfPiece(board[epsq])==Pawn);
         if (File(epsq) != 8 && board[epsq + 1] == whitePawn) {
            Square dest = epsq - 8;
            if (board[dest] == EmptyPiece)
               moves[NumMoves++] =
               CreateMove(epsq+1,dest,InvalidPiece,Pawn,Pawn,
                          EnPassant);
         }
         if (File(epsq) != 1 && board[epsq - 1] == whitePawn) {
            Square dest = epsq - 8;
            if (board[dest] == EmptyPiece)
               moves[NumMoves++] =
               CreateMove(epsq-1,dest,InvalidPiece,Pawn,Pawn,
                          EnPassant);
         }
      }
   } else {
      Bitmap pawns(board.pawn_bits[Black]);
      Bitmap pawns1 = pawns;
      pawns1.shl(7);
      pawns1.andNot(Bitmap(MAKELONGLONG(0x80808080,0x80808080)));
      pawns1.And(board.occupied[White]);
      Square dest;
      for (;(dest = pawns1.first_one()) != InvalidSquare; pawns1.clear(dest)) {
         Square loc = dest-7;
         int candidate = (Rank(loc,Black) == 7);
         if (candidate)
         {
            moves[NumMoves++] =
              CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
              CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
            if (my_ply == 0)
            {
               moves[NumMoves++] =
                 CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
               moves[NumMoves++] =
                 CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
            }
         }
         else
            moves[NumMoves++] =
              CreateMove(loc,dest,InvalidPiece,Pawn,TypeOfPiece(board[dest]),Normal);
      }
      pawns1 = pawns;
      pawns1.shl(9);
      pawns1.andNot(Bitmap(MAKELONGLONG(0x01010101,0x01010101)));
      pawns1.And(board.occupied[White]);
      for (;(dest = pawns1.first_one()) != InvalidSquare; pawns1.clear(dest)) {
         Square loc = dest-9;
         int candidate = (Rank(loc,Black) == 7);
         if (candidate)
         {
            moves[NumMoves++] =
              CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
              CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
            if (my_ply == 0)
            {
               moves[NumMoves++] =
                 CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
               moves[NumMoves++] =
                 CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
            }
         }
         else {
            moves[NumMoves++] =
              CreateMove(loc,dest,InvalidPiece,Pawn,TypeOfPiece(board[dest]),Normal);
         }
      }
      pawns1 = pawns;
      pawns1.And(Bearing::rank_mask[6]);
      pawns1.shl8();
      pawns1.andNot(board.all_occupied);
      for (;(dest = pawns1.first_one()) != InvalidSquare; pawns1.clear(dest)) {
         Square loc = dest - 8;
         moves[NumMoves++] =
            CreateMove(loc,dest,Queen,Pawn,TypeOfPiece(board[dest]),Promotion);
         moves[NumMoves++] =
            CreateMove(loc,dest,Knight,Pawn,TypeOfPiece(board[dest]),Promotion);
         if (my_ply == 0)
         {
            moves[NumMoves++] =
              CreateMove(loc,dest,Rook,Pawn,TypeOfPiece(board[dest]),Promotion);
            moves[NumMoves++] =
              CreateMove(loc,dest,Bishop,Pawn,TypeOfPiece(board[dest]),Promotion);
         }
      }
      
      if (!IsInvalid(board.EnPassantSq()))
      {
         Square epsq = board.EnPassantSq();
         ASSERT(TypeOfPiece(board[epsq])==Pawn);
         if (File(epsq) != 8 && board[epsq + 1] == blackPawn) {
            Square dest = epsq + 8;
            if (board[dest] == EmptyPiece)
               moves[NumMoves++] =
               CreateMove(epsq+1,dest,InvalidPiece,Pawn,Pawn,
                          EnPassant);
         }
         if (File(epsq) != 1 && board[epsq - 1] == blackPawn) {
            Square dest = epsq + 8;
            if (board[dest] == EmptyPiece)
               moves[NumMoves++] =
               CreateMove(epsq-1,dest,InvalidPiece,Pawn,Pawn,
                          EnPassant);
         }
      }
   }

   Bitmap pieces(board.occupied[board.Side()]);
   pieces.andNot(board.pawn_bits[board.Side()]);
   for (;(loc = pieces.first_one()) != InvalidSquare; pieces.clear(loc)) {
      int j;
      const ColorType side = PieceColor(board[loc]);
      switch (TypeOfPiece(board[loc]))
      {
      case EmptyPiece:
      case Pawn:
         break;
      case Knight:
         {
            Bitmap dests(Bearing::knight_attacks[loc]);
            if (side == White)
               dests.And(board.occupied[Black]);
            else
               dests.And(board.occupied[White]);
            Bitmap it(dests);
            Square sq;
            for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
            {
               moves[NumMoves++] = 
               CreateMove(loc,sq,InvalidPiece,Knight,
                          TypeOfPiece(board[sq]),Normal);
            }
         }
         break;
      case Bishop:
         for (j = 0; j < 4; j++)
         {
            const byte *data = BishopSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         break;
      case Rook:
         for (j = 0; j < 4; j++)
         {
            const byte *data = RookSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         break;
      case Queen:
         for (j = 0; j < 4; j++)
         {
            const byte *data = BishopSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         for (j = 0; j < 4; j++)
         {
            const byte *data = RookSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         break;
      case King:
         {
            Bitmap dests(Bearing::king_attacks[loc]);
            if (side == White)
               dests.And(board.occupied[Black]);
            else
               dests.And(board.occupied[White]);
            dests.andNot(Bearing::king_attacks[board.KingPos(
             board.OppositeSide())]);
            Bitmap it(dests);
            Square sq;
            for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
            {
               moves[NumMoves++] = 
               CreateMove(loc,sq,InvalidPiece,King,
                          TypeOfPiece(board[sq]),Normal);
            }
         }
         break;
      case InvalidPiece:
         break;
      }
   }
   moves_generated += NumMoves;
   return NumMoves;
}

void 
Move_Generator::clear_killer()
{
   for (int i = 0; i < Constants::MaxPly; i++)
   {
      Killers1[i] =
      Killers2[i] = NullMove;   
   }
}

void 
Move_Generator::clear_history()
{
   memset(HistoryW,'\0',4096*sizeof(int));
   memset(HistoryB,'\0',4096*sizeof(int));
}

void 
Move_Generator::set_killer(const Board& board, const Move & move, 
                          const unsigned ply, const unsigned depth)
{
   if (!MovesEqual(move,Killers1[ply])) {
      if (options.search.num_killers > 1)
      {
         Killers2[ply] = Killers1[ply];
      }
      if (options.search.num_killers > 0)
      {
         Killers1[ply] = move;
      }
   }
}

void 
Move_Generator::update_history(const Board& board, const Move & move, 
                              const unsigned depth)
{
   if (1) //options.search.use_history)
   {
      if (Capture(move) != EmptyPiece || PromoteTo(move) != InvalidPiece) return;
      if (board.Side() == White)
         HistoryW[(int)StartSquare(move)][(int)DestSquare(move)] += depth;
      else
         HistoryB[(int)StartSquare(move)][(int)DestSquare(move)] += depth;
   }
}

