// Copyright 1987-2003 by Jon Dart.  All Rights Reserved.
#include "bearing.h"
#include "search.h"
#include "scoring.h"
#include "constant.h"
#include "movegen.h"
#include "moveord.h"
#include "util.h"
#include "pinfo.h"
#include "attacke.h"
#include "log.h"
#include "globals.h"
#include "notation.h"
#include "pool.h"   
#include "debug.h"
#include "config.h"
#ifdef TABLEBASES
#include "tbprobe.h"
#endif

#ifdef _WIN32
static const UINT SEARCH_TIMER = 2;
#include <process.h>
#include <wincon.h>
#endif
#ifdef _TRACE
   #include <iostream>
#endif
#include <limits.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

void add_pending(char *buf);

static const int Illegal = -Constants::BIG - 1000;
static int Time_Check_Interval = 5000;

// flags for types of extension
static const int RECAPTURE = 1;
static const int CHECK = 2;
static const int PAWN_PUSH = 4;
static const int CAPTURE = 8;

Move pv[Constants::MaxPly][Constants::MaxPly];
int pv_length[Constants::MaxPly];

#ifdef _TRACE
#define UPDATE_PV(move,ply) \
{ \
   pv[ply][ply] = move; \
   for (int j=ply+1;j<ply+1+pv_length[ply+1];j++) \
     pv[ply][j] = pv[ply+1][j]; \
   pv_length[ply] = pv_length[ply+1]+1; \
   indent(ply); \
   cout << "best line[" << ply << "]: ";  \
   for (int j2 = ply; j2 < pv_length[ply]+ply; j2++) \
     cout << MoveImage(pv[ply][j2]) << " "; \
   cout << endl;  \
}
#else
#define UPDATE_PV(move,ply) \
{ \
   pv[ply][ply] = move; \
   for (int j=ply+1;j<ply+1+pv_length[ply+1];j++) \
     pv[ply][j] = pv[ply+1][j]; \
   pv_length[ply] = pv_length[ply+1]+1; \
}
#endif

Search::Search()
:ply0moves(NULL),ply0scores(NULL),age(0),
command_function(NULL),post_function(NULL),terminate_function(NULL),
 pending_function(NULL),
 hash_init_done(0),
 ratingDiff(0),
 ratingFactor(0)
{
#ifdef _WIN32
   HANDLE inh = GetStdHandle(STD_INPUT_HANDLE);
   DWORD dw;
   pipe = !GetConsoleMode(inh, &dw);
   if (!pipe)
   {
      SetConsoleMode(inh, dw & ~(ENABLE_LINE_INPUT |
                                 ENABLE_MOUSE_INPUT | ENABLE_WINDOW_INPUT));
   }
#endif   
   ply0moves = new Move[Move_Generator::MaxMoves];
   ply0scores = new int[Move_Generator::MaxMoves];
   for (int i = 0; i < Constants::MaxPly; i++)
      moves_generated[i] = new Move[Move_Generator::MaxMoves];
}

Search::~Search()
{
   cleanup(1);
}

void Search::terminate_now(int m)
{
   time_up = terminated = 1;
}

void Search::setRatingDiff(int rdiff) {
   ratingDiff = rdiff;
   ratingFactor = 0;
   if (ratingDiff < -100) {
     ratingFactor = Util::Max(PAWN_VALUE*((ratingDiff+100)/2),-PAWN_VALUE);
   }
   else if (ratingDiff > 100) {
     ratingFactor = Util::Min(PAWN_VALUE*((ratingDiff-100)/2),PAWN_VALUE);
   }
}

#ifdef _WIN32
static int is_move(const Board &b,char *buf)
{
   if (isalpha(*buf) && isdigit(buf[1]) &&
       isalpha(buf[2]) && isdigit(buf[3]))
      return 1;
   else if (strncmp(buf,"o-o",3)==0 || strncmp(buf,"O-O",3)==0 ||
            strncmp(buf,"o-o-o",5)==0 || strncmp(buf,"O-O-O",5) ==0)
      return 1;
   else
     return Notation::Value(b,b.Side(),buf) != NullMove;      
}
#endif

int Search::check_input(const Board &board)
{
#ifdef _WIN32
   DWORD nchar;
   DWORD dw;
   if (pipe)
   {
      // Use the GnuChess trick of calling PeekNamedPipe to check
      // for user input.  We need to do this because Winboard does
      // not send ^C to interrupt pondering. 
      if (PeekNamedPipe(GetStdHandle(STD_INPUT_HANDLE), NULL, 0, 
                        NULL, &nchar, NULL))
      {
         if (nchar && talkLevel == Trace) cout << "(debug) got console input nchar=" << nchar << endl;
         int nextc;
         BOOL terminate;
         for (unsigned i = 0; i < nchar; i++)
         {
            nextc=getc(stdin);
//            cout << "(debug) got: " << (char)nextc << endl;
            if (nextc == '\n')
            {
               cmd_buf[cmd_buf_count]  = '\0';
               cmd_buf_count = 0;
               // call a function (supplied by the calling code) that
               // sets terminate = 1 if we should terminate.
               command_function(cmd_buf,terminate);
               if (terminate)
               {
                  return 1;
               }
            }
            else
               cmd_buf[cmd_buf_count++] = nextc;
         }
         // Hack alert. Sometimes we get a move with no trailing
         // carriage return.
         if (cmd_buf_count > 0)
         {
            cmd_buf[cmd_buf_count] = '\0';
            if (is_move(board,cmd_buf) || (strncmp(cmd_buf,"usermove",8) == 0 && is_move(board,cmd_buf+9)))
            {
               if (talkLevel == Trace) cout << "(debug) got move w/o <CR>: " << cmd_buf << endl;
               cmd_buf_count = 0;
               command_function(cmd_buf,terminate);
               if (terminate)
               {
                  return 1;
               }
            }
         }
      }
   }
   else
   {
      // We are reading from the console, not connected to Winboard.
      // Not something we usually do, except for testing.
      int retVal = WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), 0);
      if (retVal == WAIT_OBJECT_0)
      {
         INPUT_RECORD recBuf[128];
         ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE),recBuf,
                          128,&dw);
         for (unsigned i = 0; i < dw; i++)
         {
            if (!recBuf[i].Event.KeyEvent.bKeyDown) continue;
            char nextc=recBuf[i].Event.KeyEvent.uChar.AsciiChar;
            if (nextc == (char)13 /* carriage return */)
            {
               putchar('\n');
               cmd_buf[cmd_buf_count]  = '\0';
               cmd_buf_count = 0;
               BOOL terminate;
               // call a function (supplied by the calling code) that
               // sets terminate = 1 if we should terminate.
               command_function(cmd_buf,terminate);
               if (terminate)
               {
                  return 1;
               }
               break;
            }
            else
            {
               putchar(nextc);
               cmd_buf[cmd_buf_count++] = nextc;
            }
         }
      }
   }
   return 0;
#else
   // UNIX version, uses select like Crafty does
  fd_set readfds;
  struct timeval tv;
  int data;

  FD_ZERO(&readfds);
  FD_SET(fileno(stdin), &readfds);
  tv.tv_sec=0;
  tv.tv_usec=0;
  select(16, &readfds, 0, 0, &tv);
  data=FD_ISSET(fileno(stdin), &readfds);
  int terminate = 0;
  if (data == -1)
    cerr << "warning: select returned error. errno=" << errno << endl;
  else if (data) {
     // we have something to read
     //     cout << "(debug) got input, data = " << data << endl;
     *cmd_buf = '\0'; cmd_buf_count = 0;
     char buf[256];
     memset(buf,'\0',256);
     int bytes = read(fileno(stdin),buf,255);
     if (bytes <= 0)
       return 0;
     char *p = buf;
     int stop = 0;
     while (*p) {
       char *q = cmd_buf;
       while (*p != '\n' && *p != '\0' && cmd_buf_count < 255) {
            *q++ = *p++;
            ++cmd_buf_count;
       }
       *q = '\0';
       p++; // skip over null or CR
       //cout << "(debug) cmd_buf = " << cmd_buf << endl;
       // call a function (supplied by the calling code) that
       // sets terminate = 1 if we should terminate.
       if (terminate) {
         // already terminated the search, add this cmd to
         // the stack to be executed after the search is
         // really done
         if (pending_function) pending_function(cmd_buf);
       }
       else {
          command_function(cmd_buf,stop);
          if (stop)
          {
	     terminate = 1;
          }
       }
       cmd_buf_count = 0;     
     }
     return terminate;
  }
  return(0);
#endif
}

int Search::check_time(const Board &board,int ply, int rank)
{
   if (terminated) {
      return 1;
   }
   if (command_function && check_input(board))
   {
      terminate_now();
      return 1;
   }
   if (type_of_search != Search::Fixed_Depth)
   {
      int current_time = (int)(getTime());
      // update at least the part of the stats structure that
      // is displayed in analysis mode.
      stats->elapsed_time = current_time - start_time;
      stats->num_nodes = num_nodes;
      if (current_time - (int)get_start_time() > (int)get_time_limit())
      {
         if (talkLevel == Trace) terminate_now(1);
         return 1;
      }
      if (type_of_search == Fixed_Time)
      {
         return 0;
      }
      unsigned n = theLog->current();
      
      if (xtra_time > 0 &&
          type_of_search == Time_Limit &&
          (stats->elapsed_time > time_target/2) &&
          time_added == 0) {
          // see if our score has dropped a lot since our
          // last search
          if (n >= 2 && !(*theLog)[n-2].get_book_info().is_valid() &&
             (stats->value < 5*PAWN_VALUE) &&
             (stats->value > -PAWN_VALUE*5) &&
             (stats->value < (*theLog)[n-2].score() - PAWN_VALUE/3)) {
             // search more time because our score is dropping
	         time_added++;
             if (talkLevel == Trace) {
                 cout << "(debug) adding time" << endl;
             }
          }
      }
      // Allow terminating early if this is a recapture, appeared
      // better than everything else at initial sort time, and
      // still looks good.
      if (!time_up && easy && iteration_depth >= 5) {
         // do not do this if this is the first move out of book
         if (n>=2 && !(*theLog)[n-2].get_book_info().is_valid() &&
            time_target >= 300 &&
            // Don't do this if our score is dropping ..
            (stats->value > (*theLog)[n-2].score() - PAWN_VALUE/4) &&
            // require that we have consumed at least 1/3 of the time
            // limit:
            (current_time - get_start_time())*3 > time_target) {
            if (pv[0][0] == easy_move && 
                 PieceMoved(pv[0][0]) != King) {
                time_up = 1;
                terminate_now(1);
                return 1;
            }
         }
      }
   }
   if ( time_up)
   {
      terminate_now(1);
   }
   return time_up;
}

// To reduce the chance of a hash collision, we do some elementary
// testing on moves returned by the hash table, to help weed out
// any that are invalid in the current board position ..
int hash_move_valid( const Board &board, const Move &move)
{
   if (IsNull(move))
      return 1;
   if (StartSquare(move) == DestSquare(move))
      return 0;
   Piece p = board[StartSquare(move)];
   if (TypeOfPiece(p) != EmptyPiece && PieceColor(p) == board.Side())
   {
      p = board[DestSquare(move)];
      if (IsEmptyPiece(p) || PieceColor(p) != board.Side())
      {
         if (TypeOfMove(move) == EnPassant)
         {
            return board.EnPassantSq() != InvalidSquare;
         }
         else if (TypeOfMove(move) == KCastle)
         {
            return (board.CastleStatus(board.Side()) ==
                    Board::CanCastleKSide) ||
            (board.CastleStatus(board.Side()) ==
             Board::CanCastleEitherSide);
         }
         else if (TypeOfMove(move) == QCastle)
         {
            return (board.CastleStatus(board.Side()) ==
                    Board::CanCastleQSide) ||
            (board.CastleStatus(board.Side()) ==
             Board::CanCastleEitherSide);
         }
         else
            return 1;
      }
      else
         return 0;
   }
   else
      return 0;
}

void Search::show_status(const Board &board, const char ch)
{
   if (terminated)
      return;
   int ply = stats->depth;
   static char move_text[20];
   Notation::Image(board, pv[0][0]/*move*/, move_text);
   const char *val = Scoring::print_score(stats->display_value);
   if (talkLevel == Whisper)
   {
      // Arasan is now more silent. Change the following line to
      // whisper during a search.
      if (0) //stats->elapsed_time >= 300 && time_target > 500)
      {
         cout << "tellics whisper ply " << ply;
         if (ply < 10) cout << ".  ";
         else cout << ". ";
         cout << move_text << " score: " << val << " " <<
         "time: " << stats->elapsed_time/100 << " secs. nodes: " <<
         stats->num_nodes << endl;
      }
   }
   else if (talkLevel == Debug)
   {
      cout << "ply " << ply;
      if (ply < 10) cout << ".  ";
      else cout << ". ";
      cout << move_text << '\t' << stats->elapsed_time/100 << 
      " seconds.\tscore: " << val  << "\t" <<
      stats->num_nodes << " nodes." << endl;
   }
   if (!is_background_search())
   {
      if (post_function) post_function(*stats);
   }
}

void Search::set_background(int b)
{
   bkgrnd_search = b;
   if (!b)
   {
      // changing from background to foreground search,
      // update display
      show_status(initial_board,0);
   }
}

int Search::is_draw( const Board &board, int &rep_count, int ply)
{
   // First check for draw by repetition
   int entries = board.move_count-2;
   hash_t to_match = board.HashCode();
   rep_count = 0;
   if (entries >= 1) {
      for (hash_t *rep_list=board.rep_list_head-3;
          entries>=0;
          rep_list-=2,entries-=2)
      {
         if (*rep_list == to_match)
         {
            rep_count++;
            if (rep_count >= 2 ||
                (rep_count == 1 && board.rep_list_head-rep_list <= ply)) {
               return 1;
            }
   
         }
      }
   } 
   // check the 50 move rule
   if (board.move_count >= 100) {
      return 1;
   }
   // check for insufficient material and other drawing situations
   return Scoring::material_draw(board) || Scoring::is_theoretical_draw(board);
}

int Search::evalu8(const Board & board, int alpha, int beta, int ply)
{
   // Evaluate a board position. 
   int rep_count;
   if (is_draw(board,rep_count,ply))
      return draw_score(board);
   int score = Scoring::material_score(board);
   int pos_score = Scoring::positional_score(board,-Constants::BIG,
                                             Constants::BIG);
   score += pos_score;
   return score;
}

int Search::draw_score(const Board &board)
{
   int score = 0;

   // A draw is worse if there is a lot of material on the board
   if (root_total_mat > 6000)
   {
      score -= 16;
   }
   else if (root_total_mat > 4800)
   {
      score -= 16*(root_total_mat-4800)/1200;
   }
   if (board.Side() != side_to_move)
   {
      score = -score;
   }
   int mat_diff = Scoring::material_score(board);
   // Penalize the side to move for allowing a draw if it is ahead in
   // material
   const Material &my_mat = board.getMaterial(board.Side());
   if (Util::Abs(mat_diff) >= 128 &&
       (!my_mat.no_pawns() || my_mat.value() > 2320))
   {
      if (mat_diff > 0)
         score -= (mat_diff-128)/8;
      else
         score -= (mat_diff+128)/8;
   }
   // if we know the opponent's rating (which will be the case if playing
   // on ICC, factor that into the draw score - a draw against a high-rated
   // opponent is good; a draw against a lower-rated one is bad.
   if (ratingDiff != 0) {
      if (board.Side() == side_to_move)
        score += ratingFactor;
      else
        score -= ratingFactor;
   }
#ifdef _TRACE
   cout << "(debug) draw score=" << score << endl;
#endif
   return score;
}

int Search::
ply0_move_search(Board & board, int alpha, int beta,
                 int depth,
                 int &terminate)
{
   // implements alpha/beta search for the top most ply.  We use
   // the PVS (principal variation search) variant of alpha/beta.

   time_check_counter++;
   num_nodes++;

   int in_pv = 1;
   int in_check = 0;
   extensions[0] = 0;
   checks[0] = 0;
   int move_count = 0;
   int move_index;
   int num_try = 0; // # of legal moves tried
   current_score = alpha;

   int initial_alpha = alpha;

   int score = alpha;

   // At this point we need to know if we are in check or not.
   in_check = checks[0] = board.CheckStatus() == Board::InCheck;
   Board_State save_state(board);

   int try_score;
   //
   // re-sort the ply 0 moves.
   // Pv first, then captures, then killers. Killers at this ply (ply 0)
   // are moves that have previously exceeded alpha.
   //
   if (iteration_depth>0) Move_Ordering::reorder_ply0_moves(board,ply0moves, ply0move_count, ply0first);
   //
   // Search the next ply
   //
   int newbeta = beta;
   move_index = 0;
   while (move_index < ply0move_count)
   {
      stats->mvleft = stats->mvtot-move_index;
      Move move = ply0moves[move_index++];
      if (IsMate(move)) continue; // move leads to mate, skip it
#ifdef _TRACE
      cout << "window = [" << alpha << ',' << newbeta << ']' <<
      " beta = " << beta << endl;
      cout << "trying " << 0. << ". " << MoveImage(move) << endl;
#endif
      last_move[0] = move;
      board.DoMove(move);
      int extend = 0;
      ++num_try;
#ifdef TABLEBASES  
  if (options.search.use_tablebases) {
     const Material &wMat = board.getMaterial(White);
     const Material &bMat = board.getMaterial(Black);
     int tb_hit = 0;
     int tb_score;
     if (wMat.piece_count() + bMat.piece_count() <= EGTBMenCount) {
      stats->tb_probes++;
      tb_hit = probe_tb(board, tb_score, 0);
     } else tb_hit = 0;
     if (tb_hit) {
#ifdef _TRACE
       cout << "EGTB hit: score " << tb_score << endl;
#endif
        stats->tb_hits++;
        board.UndoMove(move,save_state);
        try_score = -tb_score;
        goto tb_shortcut;
     }
  }
#endif

      extend = 0;
      if (options.search.pawn_push_extensions && 
          (PieceMoved(move) == Pawn) && 
          (Rank(DestSquare(move),board.OppositeSide()) == 7))// &&
	  //          (depth > 2*DEPTH_INCREMENT || !board.any_attacks(DestSquare(move),board.Side())))
      {
         extend += PAWN_PUSH_EXTENSION;
         extensions[0] |= PAWN_PUSH;
         stats->pawn_extensions++;
      }
      if (options.search.check_extensions && in_check)
      {
         extensions[0] |= CHECK;
         extend += CHECK_EXTENSION;
         stats->check_extensions++;
      }
      if (extend >DEPTH_INCREMENT) extend = DEPTH_INCREMENT;
      if (depth+extend-DEPTH_INCREMENT >= 0)
      {
         try_score = -move_search(board, -newbeta, -alpha,
                                  1, depth+extend-DEPTH_INCREMENT,
                                  terminate);
      }
      else
      {
         try_score = -quiesce(board, -newbeta, -alpha,
                              1, 0);
      }
      if (terminate)
      {
         board.UndoMove(move,save_state);
         current_score = alpha;
         return alpha;
      }
      // We weeded out illegal moves so should never get one here
      ASSERT(try_score != Illegal);
#ifdef _TRACE
      cout << "0. " << MoveImage(move) << ' ';
      cout << try_score;
      if (in_pv) cout << " (pv)";
      cout << endl;
#endif
      if (!in_pv &&
          ((try_score > alpha) && (try_score < beta)))
      {
         // We failed to get a cutoff and must re-search
#ifdef _TRACE
         cout << "window = [" << alpha << "," << beta
         << "]" << endl;
         cout << "score = " << try_score << " - no cutoff, researching .." << endl;
#endif
         if (depth+extend -DEPTH_INCREMENT >= 0)
            try_score=-move_search(board,
                                   -beta,-alpha,1,depth+extend-DEPTH_INCREMENT,terminate);
         else
            try_score=-quiesce(board,
                               -beta,-alpha,1,0);
#ifdef _TRACE
         cout << "0. " << MoveImage(move) << ' ';
         cout << try_score;
         cout << endl;
#endif
      }

      board.UndoMove(move,save_state);
      if (terminate)
      {
         current_score = alpha;
         return alpha;
      }
#ifdef TABLEBASES
   tb_shortcut:
#endif
      ply0scores[move_index-1] = try_score;
      // If this move leads to being mated, remove it from the
      // ply 0 move list - there is no use searching it again
      if (try_score <= -Constants::BIG+(int)iteration_depth)
      {
#ifdef _TRACE
         cout << "removing move " << MoveImage(move) << endl;
#endif
         for (int i = 0; i < ply0move_count; i++)
         {
            if (MovesEqual(ply0moves[i],move))
            {
               SetMate(ply0moves[i]);
               break;
            }
         }

      }
      if (try_score > alpha)
      {
         current_score = score = try_score;
         Move_Generator::set_killer(board,last_move[0], 0, depth);
         if (try_score >= beta)
         {
            best[0] = move;
#ifdef _TRACE
			cout << "ply 0 beta cutoff" << endl;
#endif
            if (pv)
            {
               pv[0][0] = move;
               pv_length[0] = 1;
            }
            return try_score;
         }
         int pv_change = !MovesEqual(best[0],move);
         best[0] = move;
         if (pv_change)
         {
#ifdef _TRACE1
	   cout << "pv_change" << endl;
#endif
            UPDATE_PV(move,0);
            update_stats(alpha,beta);
            show_status(board,'>');
         }
         alpha = try_score;
      }
      else if (in_pv && try_score < initial_alpha)
      {
         // we have failed low and must reset the window
         best[0] = move;
#ifdef _TRACE
         cout << "ply 0 alpha cutoff" << endl;
#endif
         current_score = try_score;
         return try_score;
      }
      in_pv = 0;
      newbeta = alpha + 1; // zero-width window
   }

   if (num_try == 0 && move_count == 0)
   {
      // no moves were tried
      if (in_check)
      {
         if (ply0move_count == 0)        // mate
         {
            score = -(Constants::BIG);
            state = Checkmate;
         }
         else
         {
            // We generated some moves, and some were legal,
            // but we skipped them all.
            score = evalu8(board,alpha,beta,0);
         }
      }
      else    // stalemate
      {
         state = Stalemate;
#ifdef _TRACE
         cout << "stalemate!" << endl;
#endif
         score = draw_score(board);
      }
   }
   else if (num_try > 0 && score < beta)
   {
       // If there's only one legal move, don't search any more:           
       if (num_try == 1 && move_index == move_count)
          terminate = 1;
   }
   if (alpha != initial_alpha && score < beta)
   {
      if (num_try == 0)
         pv_length[0] = 0;
      else
      {
         UPDATE_PV(best[0],0);
      }
   }
   ASSERT(score >= -Constants::BIG && score <= Constants::BIG);
   return score;
}

int Search::calc_extensions(const Board &board,
                            int ply,int depth, int in_check, int forced,
                            int threat) 
{
   int extend = 0;
   // see if we should apply any extensions at this node.
   extensions[ply] = 0;
   int limit;
   if (iteration_depth < 10)
      limit = (ply > iteration_depth*2);
   else
      limit = (ply > iteration_depth + 10);
   if (options.search.forced_extensions && checks[ply] && forced) {
       // one single reply to check
       extend += FORCED_EXTENSION;
       stats->forced_extensions++;
   }
   if (options.search.check_extensions && in_check)
   {
       extensions[ply] |= CHECK;
       extend += CHECK_EXTENSION;
       stats->check_extensions++;
   }
   if (options.search.pawn_push_extensions && !IsNull(last_move[ply-1]) && 
       (PieceMoved(last_move[ply]) == Pawn) &&
       (Rank(DestSquare(last_move[ply]),board.OppositeSide()) == 7)) // &&
       //(depth > 2*DEPTH_INCREMENT || !board.any_attacks(DestSquare(last_move[ply]),board.Side())))
   {
      extend += PAWN_PUSH_EXTENSION;
      extensions[ply] |= PAWN_PUSH;
      stats->pawn_extensions++;
   }
   if (options.search.recapture_extensions && 
       Capture(last_move[ply]) != EmptyPiece &&
       !(extensions[ply-1] & RECAPTURE) &&
       Capture(last_move[ply-1]) != EmptyPiece &&
       Capture(last_move[ply-1]) != Pawn &&
       DestSquare(last_move[ply-1]) == DestSquare(last_move[ply]))
   {
      int capVal = PieceValue(Capture(last_move[ply-1]));
      int capVal2 = PieceValue(Capture(last_move[ply]));
      if( Util::Abs(capVal - capVal2) < PAWN_VALUE/2)
      {
         extend += RECAPTURE_EXTENSION;
         extensions[ply] |= RECAPTURE;
         stats->recap_extensions++;
      }
   }
   if (options.search.capture_extensions && 
       (ply >=2 && Capture(last_move[ply]) != EmptyPiece) &&
       Capture(last_move[ply]) != EmptyPiece &&
       !(extensions[ply-1] & RECAPTURE) &&
       (PieceValue(Capture(last_move[ply]))> PAWN_VALUE) &&
       (depth > (iteration_depth*DEPTH_INCREMENT)/2)) {
     extend += DEPTH_INCREMENT/2;
     extensions[ply] |= RECAPTURE_EXTENSION;
     stats->cap_extensions++;
   }

   if (options.search.threat_extensions && threat) {
       stats->threat_extensions++;
       extend += threat;
   }
   if (limit) {
       extend /= 2;
   }
   if (extend > DEPTH_INCREMENT) extend = DEPTH_INCREMENT;
   return extend;
}

int Search::
quiesce(Board & board, int alpha, int beta,
        const int ply, int depth)
{
   // recursive function, implements quiescence search.
   //
   ASSERT(ply < Constants::MaxPly);
   num_nodes++; time_check_counter++; num_qnodes++;
   pv_length[ply] = 0;
   int score;
   if (ply >= Constants::MaxPly-1)
   {
      return evalu8(board,alpha,beta,ply);
   }
#ifdef _TRACE
   indent(ply); cout << "window [" << alpha << "," << beta << "]" << endl;
#endif 
   int stand_pat_score;
   int mat_score = -Constants::BIG;

/**   if (is_draw(board,rep_count,ply)) {
      score = draw_score(board);
      if (score > alpha && score < beta)
      {
          pv[ply][ply] = NullMove;
      }
      return score;
   }
   else **/ {
      // Establish a default score.  This score is returned if no
      // captures are generated, or if no captures generate a better
      // score (since we generally can choose whether or not to capture).
      mat_score = Scoring::material_score(board);
      if (mat_score - 5*PAWN_VALUE > beta) {
#ifdef _TRACE
         indent(ply); cout << "** CUTOFF ** (material)" << endl;; 
#endif             
         return mat_score;
      }
      else
         stand_pat_score = mat_score+Scoring::positional_score(board,alpha-3*PAWN_VALUE-mat_score,beta-mat_score);
   }
   
#ifdef _TRACE
   indent(ply);
   cout << "cap score = " << stand_pat_score << endl;
#endif
   if (stand_pat_score > alpha)
   {
      if (stand_pat_score >= beta)
      {
#ifdef _TRACE
         indent(ply); cout << "** CUTOFF **" << endl;; 
#endif             
         return stand_pat_score;
      }
      alpha = stand_pat_score;
   }
   score = alpha;
   Move *moves = moves_generated[ply];
   Move_Generator mg(board, ply, NullMove);
   // generate all the capture moves
   int move_count = mg.Generate_Captures(moves);
#ifdef _TRACE
   indent(ply); cout << move_count << " capture moves, ";
#endif
   if (move_count == 0) {
#ifdef _TRACE
      cout << "0 after pruning" << endl;
#endif
      return score; // no capture moves
   }
   num_moves += move_count;
   // order captures (and prune losing captures)
   move_count = Move_Ordering::order_captures(board, moves, move_count, alpha,
                                              stand_pat_score,ply);
#ifdef _TRACE
   cout << move_count << " after pruning." << endl;
#endif
   if (move_count == 0) return score;

   Board_State save_state(board);

   int initial_alpha = alpha;
   int move_index = 0;
   int num_try = 0; // # of legal moves tried
   int try_score;
   best[ply] = NullMove;

   const int swag = QSEARCH_FORWARD_PRUNE_BIAS+stand_pat_score;
   while (move_index < move_count)
   {
      Move emove = moves[move_index++];
      if (Capture(emove) == King)
         return -Illegal;
      // We already did forward pruning in order_captures, but if alpha has
      // increased we have an opportunity to do it again.
      if (alpha != initial_alpha)
      {
         int est = PieceValue(Capture(emove));
         if (PromoteTo(emove) != InvalidPiece)
            est += (int)PieceValue(PromoteTo(emove)) - PAWN_VALUE;
         if (swag + est < alpha)
         {
            continue;
         }
         else
         {
            est = est - PieceValue(PieceMoved(emove));
            if (est <= 10 && Capture(emove) != EmptyPiece) 
            {
               est = attack_estimate(board,emove);
               if (swag + est < alpha) {
                  continue;
               }
            }
         }
     }
#ifdef _TRACE
      indent(ply);
      cout << "trying " << ply << ". " << MoveImage(emove) << endl;
#endif
      last_move[ply] = emove;
      board.DoMoveQ(emove);
      try_score = -quiesce(board, -beta, -alpha, ply+1, depth-1);
      board.UndoMove(emove,save_state);
      if (try_score != Illegal)
      {
#ifdef _TRACE
         indent(ply);
         cout << ply << ". " << MoveImage(emove) << ' ';
         cout << try_score;
         cout << endl;
#endif
         num_try++;
         if (try_score > alpha)
         {
            score = try_score;
            if (try_score >= beta)
               break;
            best[ply] = emove;
            alpha = try_score;
            if (score >= Constants::BIG-1-ply)
               break; // mating move found
         }
      }
   }
   if (score >= beta)
   {
#ifdef _TRACE
      indent(ply);
      cout << "**CUTOFF**" << endl;
#endif 
      return score;
   }
   if (alpha != initial_alpha && score < beta)
   {
      if (IsNull(best[ply]))
         pv_length[ply] = 0;
      else
      {
         UPDATE_PV(best[ply],ply);  
         //Move_Ordering::update_history(board,best[ply],depth);
      }
   }
   return score;
}

Move Search::find_best_move(
                           const Board &board, 
                           Search::Search_Type srcType,
                           int32 timelimit,
                           int32 xtratime,
                           int plylimit,
                           const int background,
                           Statistics &stat_buf,
                           Talk_Level t)
{
   talkLevel = t;
   time_target = timelimit;
   xtra_time = xtratime;
   ply_limit = plylimit;
   type_of_search = srcType;
   stats = &stat_buf;
   stats->clear();

   bkgrnd_search = background;
   Board board_copy(board);
   initial_board = board;

   init(board_copy);

   if (Scoring::is_legal_draw(board))
   {
      // If it's a legal draw situation before we even move, then
      // just return a draw score and don't search.
      stats->state = Draw;
      stats->value = draw_score(board);
   }
   else {
      // Do the actual search
      iterate(board_copy);
   }

   static const int end_of_game[] = { 0, 0, 0, 1, 1,
      1, 1};
   stats->end_of_game = end_of_game[(int)stats->state];
   if (!stats->end_of_game && options.search.can_resign)
   {
      if (stats->depth >= 2)
      {
         // Use the display value - i.e. do not use score if
         // we are failing low
         if (stats->display_value <= 8-(int)Constants::BIG)
         {
            // TBD we can sometimes get -10000 as a score!
            // this is bogus, we should not resign.
            if (stats->display_value != -(int)Constants::BIG) 
               stats->state = Resigns;
         }
         else if (stats->display_value > 100-Constants::BIG &&
		  ((100*stats->display_value)/PAWN_VALUE <= options.search.resign_threshold))
             stats->state = Resigns;
         stats->end_of_game = end_of_game[(int)stats->state];
      }
   }

   if (talkLevel == Debug)
   {
      if (stats->elapsed_time > 0)
         cout << (100.0*(float)stats->num_nodes)/((float)stats->elapsed_time) << " nodes/second." << endl;
      cout << (num_nodes-num_qnodes) << " regular nodes, " <<
      num_qnodes << " quiescence nodes." << endl;
      cout << hash_searches << " searches of hash table, " <<
 	  hash_hits << " successful";
      if (hash_searches != 0) 
        cout << " (" << 
        (int)((100.0*(float)hash_hits)/((float)hash_searches)) <<
        " percent).";
      cout << endl;
      cout << hash_inserts << " hash entries inserted, " <<
      hash_replaces << " entries replaced, " <<
      hash_inserts_failed << " inserts failed." << endl;
      cout << "move ordering: ";
      static const char *labels[] = {"1st","2nd","3rd","4th"};
      for (int i = 0; i < 4; i++) {
         cout << labels[i] << " " << 
         (100.0*stats->move_order[i])/(float)stats->move_order_count << "% " ;
      }
      cout << endl;
      cout <<  "null cutoffs: " << stats->num_null_cuts << " (" <<
              (int)((100.0*(float)stats->num_null_cuts)/((float)stats->num_nodes)) <<
              "%)" << endl;
      cout << "pruning: " << 
#ifdef FUTILITY_PRUNE
              stats->futility_pruning << " moves (futility)" <<
#endif
#ifdef RAZORING
#ifdef FUTILITY_PRUNE
              ", " << 
#endif
              stats->num_razored << " nodes (razored)" << 
#endif
              endl;
      cout << "extensions: " << 
              stats->check_extensions << " check, " <<
              stats->recap_extensions << " recapture, " <<
              stats->cap_extensions << " capture," << endl;
      cout << "            " <<   
              stats->pawn_extensions << " pawn, " << 
              stats->threat_extensions << " threat, " << 
              stats->forced_extensions << " forced move." << endl;
      cout << stats->tb_probes << " tablebase probes, " <<
              stats->tb_hits << " tablebase hits" << endl;
      fflush(stdout);
   }
   return pv[0][0];
}

void Search::update_stats(int alpha, int beta)
{
   time_t end_time = getTime();
   stats->elapsed_time = end_time - start_time;
   stats->num_nodes = num_nodes;
   stats->num_qnodes = num_qnodes;
   stats->num_moves = num_moves;
   // Note: the search routine counts depth internally starting
   // at 0, but we report it starting at 1 (as most other programs do).
   stats->depth = iteration_depth+1;
   stats->value = current_score;
   // Leave the previous value intact if we have failed low.
   if (!(stats->value == alpha && iteration_depth > 0))
   {
      stats->display_value = stats->value;
   }
   stats->state = state;
   if (terminated)
   {
      // keep the previous "best line" intact.
      return;
   }
   *(stats->best_line_image) = '\0';
   stats->best_line[0] = NullMove;

   Board board_copy(initial_board);
 
   if (IsNull(pv[0][0]))
   {
#ifdef _TRACE
       cout << "warning: pv is null\n";
#endif       
       return;
   }
   else if (pv_length[0] == 0)
   {
       return;
   }
   int i = 0;
   char *p = stats->best_line_image;
   while (i < pv_length[0] && !IsNull(pv[0][i]))
   {
      stats->best_line[i] = pv[0][i];
      char move_image[20];
      *move_image = '\0';
      Notation::Image(board_copy,stats->best_line[i],move_image);
      int len = strlen(move_image);
      if (p+len+1-stats->best_line_image > 255)
	break;
      strcpy(p,move_image);
      p+= len;
      strcpy(p," ");
      p++;
      board_copy.DoMove(stats->best_line[i]);
      ++i;
      int rep_count;
      if (is_draw(board_copy,rep_count,0))
      {
         break;
      }
      if (pv_length[0] < 2 && use_hash_table)
      {
         // get the next move from the hash table, if possible
         // (for pondering)
         Position_Info *hash_entry = searchHash(board_copy,rep_count);
         if (hash_entry)
         {
            Move hashMove = hash_entry->best_move(board_copy);
            if (!IsNull(hashMove))
            {
               stats->best_line[i] = hashMove;
               *p = '\0';
               Notation::Image(board_copy,hashMove,move_image);
               int len = strlen(move_image);
               if (p+len+1-stats->best_line_image > 255)
                 break;
               strcpy(p,move_image);
               p+=len; 
               ++i;
            }
         }
         break;
      }
   }
   *p = '\0';

   stats->best_line[i] = NullMove;
#ifdef _TRACE
   cout << "best line: " << stats->best_line_image << endl;
#endif
}

void Search::iterate(const Board &board)
{
   Time_Check_Interval = 5000;

   int terminate = 0;
   if (ply_limit > 0)
      ply_limit = Util::Min(ply_limit-1,Constants::MaxPly-1);

   // Generate the ply 0 moves here:
   Move_Generator mg(board,0,NullMove);
   ply0move_count = mg.GenerateAllMoves(ply0moves,0);
   num_moves += ply0move_count;
   // Searching may alter the board, so operate on a copy:
   Board board_copy(board);
   // Order the moves.  The ordering process removes any illegal moves
   // at ply 0, so fix the move count if necessary.
   stats->mvtot = ply0move_count = 
   Move_Ordering::order_ply0_moves(board_copy, ply0moves, ply0move_count, 
                                   NullMove,
                                   easy);
   // See if there is an "easy" move (a recapture that appears to
   // gain more material than any other alternative).  If so we
   // may be able to terminate the search early.
   easy &= (ply0move_count > 0 &&
            board[DestSquare(ply0moves[0])] != EmptyPiece &&
            game_moves->num_moves() > 0 &&
	        (DestSquare(game_moves->move(game_moves->num_moves()-1)) ==
            DestSquare(ply0moves[0])) &&
            (Capture(game_moves->move(game_moves->num_moves()-1)) != EmptyPiece &&
            Capture(ply0moves[0]) != EmptyPiece)
            );
   easy_move = ply0moves[0];
   // Positions are stored in the hashtable with an "age" to identify
   // which search they came from.  "Newer" positions can replace 
   // "older" ones.  Update the age here since we are starting a
   // new search.
   age = (age + 1) % 3;
   int rep_count;
   int drawAtRoot = is_draw(board,rep_count,0);
   // Incrementally search the board to greater depths, up to the
   // ply limit
   for (iteration_depth = 0;
       iteration_depth <= ply_limit && !terminate && !time_up; 
       iteration_depth++)
   {
      int value;
      int lo_window, hi_window;
      if (iteration_depth == 0)
         current_score = value = Scoring::evalu8(board);
      else {
         value = stats->value;
         if (drawAtRoot) break;
      }
      lo_window = Util::Max(-Constants::BIG,value - INITIAL_SEARCH_WINDOW / 2);
      hi_window = Util::Min(value + INITIAL_SEARCH_WINDOW / 2,
                            Constants::BIG);
  int tb_hit = 0;
#ifdef TABLEBASES
  if (options.search.use_tablebases) {
     const Material &wMat = board.getMaterial(White);
     const Material &bMat = board.getMaterial(Black);
     int tb_score;
     // Like Crafty, we restrict tablebase probes to near the top
     // of the tree, unless a promotion or capture is made.
     if(wMat.piece_count() + bMat.piece_count() <= EGTBMenCount) {
       stats->tb_probes++;
       tb_hit = probe_tb(board, tb_score, 0);
     }
  }
#endif
#ifdef _TRACE
      cout << "iteration " << iteration_depth << " window = [" <<
        lo_window << "," << hi_window << "]" << endl;
#endif
      int failhigh = 0;
      int faillow = 0;
      while (1)
      {
         if (!failhigh && !faillow)
            ply0first = best[0];
         ply0_move_search(board_copy, lo_window, hi_window, 
                          DEPTH_INCREMENT*iteration_depth, terminate);
         update_stats(lo_window,hi_window);
         if (iteration_depth >= 2) {
            if (ply0move_count == 1 ||
              Scoring::is_legal_draw(board))
            {
              // Draw, or there is only one legal move, terminate.
              terminate = 1;
            }
            else if (tb_hit) terminate = 1;
         }
         else if (state == Checkmate || state == Stalemate) {
            terminate = 1;
            break;
         }
         if (stats->elapsed_time > 100)
         {
            long denom = bkgrnd_search ? 200L : 500L;
            Time_Check_Interval = 
              (int)(denom*stats->num_nodes)/(36L*stats->elapsed_time);
         }
         if (!terminate)
         {
            terminate = check_time(board,0,0);
            if (terminate_function && terminate_function(*stats))
            {
               terminate = 1;
            }
         }
         if (terminate)
         {
            show_status(board,0);
            break;
         }
         else
         {
            show_status(board,0);
            ASSERT(stats->value != Constants::BIG);
            if (stats->value < hi_window && stats->value > lo_window) {
                // score is in the bounds of the search window,
                // no need to re-search
                break;
            }
            if (stats->value >= hi_window)
            {
               // shouldn't really happen, but put a check here.
               // Score may reach +Constants::BIG sometimes, indicating
               // a mating move (shouldn't be >Constants::BIG-1).
               if (failhigh) break;
#ifdef _TRACE
               cout << "ply 0 high cutoff, re-searching ..." << endl;
#endif
               // high cutoff, must re-search
#ifdef _TRACE
               cout << "last_move[0] = " << MoveImage(last_move[0]) << endl;
#endif
               ply0first = last_move[0]; // so we research this move
               hi_window = Constants::BIG;
               lo_window = Util::Min(stats->value,Constants::BIG-Constants::MaxPly);
               failhigh = 1;
               faillow = 0;
            }
            else if (stats->value <= lo_window && !failhigh)
            {
#ifdef _TRACE
               cout << "ply 0 low cutoff, re-searching ..." << endl;
               cout << "value = " << stats->value << endl;
#endif
               hi_window = lo_window;
               lo_window = -Constants::BIG;

               failhigh = 0;
               faillow = 1;
            }
            else
            {
               // Use infinite window - next search score should be in 
               // bounds.
               hi_window = Constants::BIG;
               lo_window = -Constants::BIG;
               failhigh = faillow = 0;
            }
         }
         best_moves[iteration_depth] = best[0];
         best_scores[iteration_depth] = stats->value;
      }
#ifdef _TRACE
      cout << iteration_depth+1 << " ply search result: " << MoveImage(pv[0][0]) << 
      "   value = " << stats->value << endl;
#endif
      if (stats->value < (int)(iteration_depth*2) - Constants::BIG)
      {
         // We're either checkmated or we certainly will be, so
         // quit searching.
#ifdef _TRACE
         cout << "terminating, low score" << endl;
#endif            
         break;
      }
      else if (stats->value >= (Constants::BIG - (int)(iteration_depth*2)
                                  - 1))
      {
#ifdef _TRACE
         cout << "terminating, mate score" << endl;
#endif            
         break;
      }
   }
}


int Search::get_ply() const
{
   // return current depth of search.
   return iteration_depth;
}

unsigned long Search::get_numnodes() const
{
   // return # of nodes visited
   return num_nodes;
}    

void Search::init(const Board &board)
{
   num_nodes = num_qnodes = num_moves = hash_hits = 
               hash_searches = hash_inserts = hash_replaces = hash_inserts_failed = 0L;
   time_check_counter = 0;
   time_up = 0;
   terminated = 0;
   time_added = 0;
   iteration_depth = 0;
   current_score = 0;
   state = NormalState;
   side_to_move = board.Side();
   root_total_mat = board.getMaterial(White).value() + board.getMaterial(Black).value();
   start_time = getTime();
   for (int i = 0; i< Constants::MaxPly;i++)
   {
      pv[i][0] = best[i] = last_move[i] = best_moves[i] = NullMove;
      extensions[i] = checks[i] = 0;
      pv_length[i] = 0;
   }
   ply0first = NullMove;
   Scoring::init(board);
   root_score = Scoring::evalu8(board);

   Move_Generator::clear_killer();
   Move_Generator::clear_history();

   null_depth = options.search.null_depth_reduction*DEPTH_INCREMENT;

   // Allocate the hash table.  Size is in kilobytes.  Guesstimate
   // a conversion to the number of hash table entries. 
   use_hash_table = options.search.hash_table_size > 0;
   int hash_entries;
   hash_entries = (1000*options.search.hash_table_size)/sizeof(Position_Info);
   if (hash_init_done && hash_table_size != options.search.hash_table_size) {
      // hash table was resized (GUI can do this)
      reinit();
   }
   if (!hash_init_done && use_hash_table)
   {
      initHash(hash_entries);
      hash_table_size = options.search.hash_table_size;
      hash_init_done = 1;
   }
}

void Search::reinit()
{
   // After the search options change, we may need to reallocate
   // the hash table.  Free it here so that next time through init()
   // the reallocation will be done.
   freeHash();
   hash_init_done = 0;
}

long Search::calc_hash_size()
{
   // Size the hash table according to how much memory we have.
   static int sizes[12]={ 997, 1997, 3001, 4003, 4999, 6007, 6997, 8003,
      10007, 12007, 15013, 18103};

   int i = 0;
   return 32L*sizes[i];
}

void Search::cleanup(int clearHash)
{
   if (hash_init_done && clearHash)
   {
      freeHash();
   }
   delete [] ply0moves;
   delete [] ply0scores;
   for (int i = 0; i< Constants::MaxPly;i++)
   {
      if (moves_generated[i] != NULL)
         delete [] moves_generated[i];
   }
}
